import json

import re
from .basic_device import BasicDevice, DeviceError
from datetime import datetime, timedelta
from .const import ENDPOINT_HH, COMMAND_GET_PROGRAM, COMMAND_GET_COMMAND
import locale

locale.setlocale(locale.LC_ALL, '')  # Use '' for auto, or force e.g. to 'en_US.UTF-8'

COMMAND_VALUE_ECOM_STAT_TOTAL = 'ecomXstatXtotal'
COMMAND_VALUE_ECOM_STAT_AVG = 'ecomXstatXavarage'

REGEX_MATCH_LITER = r"(\d+(?:[\,\.]\d+)?).?ℓ"
REGEX_MATCH_KWH = r"(\d+(?:[\,\.]\d+)?).?kWh"


class WashingMachine(BasicDevice):
    """Class representing V-Zug washing machines"""

    def __init__(self, host: str, username: str = "", password: str = ""):
        super().__init__(host, username, password)
        self._seconds_to_end = 0
        self._program_name = ""
        self._program_status = ""
        self._optidos_active = False
        self._optidos_config = ""
        self._optidos_a_status = ""
        self._optidos_b_status = ""
        self._power_consumption_kwh_total = 0.0
        self._water_consumption_l_total = 0.0
        self._power_consumption_kwh_avg = 0.0
        self._water_consumption_l_avg = 0.0

    def _reset_program_information(self) -> None:
        self._seconds_to_end = 0
        self._program_name = ""
        self._program_status = ""
        self._optidos_active = False
        self._optidos_config = ""
        self._optidos_a_status = ""
        self._optidos_b_status = ""

    async def load_all_information(self) -> bool:
        """If a program is active load additional program details"""
        loaded = await super().load_all_information()
        if loaded and self.is_active:
            loaded = await self.load_program_details()

        return loaded

    async def load_program_details(self) -> bool:
        """Load program details information by calling the corresponding API endpoint"""

        self._logger.info("Loading program information for %s", self._host)

        self._reset_program_information()
        try:
            program_json = (await self.make_vzug_device_call_json(
                self.get_command_url(ENDPOINT_HH, COMMAND_GET_PROGRAM)))[0]

            self._program_status = program_json['status']

            # Load optiDos detailed information if optiDos is available / active
            # (optiDos may be available even if no program is active...)
            self._read_optidos_details(program_json)

            if 'active' not in self._program_status:
                self._logger.info("No program information available because no program is active")
                return False

            self._program_name = program_json['name']
            self._seconds_to_end = program_json['duration']['act']

            self._logger.info("Go program information. Active program: %s, minutes to end: %.0f, end time: %s",
                              self.program_name, self.seconds_to_end / 60, self.date_time_end)

            return True

        except DeviceError as e:
            self._error_code = e.error_code
            self._error_message = e.message
            self._error_exception = e
            return False

    def _read_optidos_details(self, program_json: json) -> None:
        """Read optiDos information from given program response"""

        if 'fillLevelA' in program_json:
            self._optidos_a_status = program_json['fillLevelA']['act']

        if 'fillLevelB' in program_json:
            self._optidos_b_status = program_json['fillLevelB']['act']

        self._optidos_active = False
        if 'optiDos' in program_json:
            self._optidos_config = program_json['optiDos']['set']

            # TODO: Add support for other optiDos configurations
            if self._optidos_config == "detergentAandB":
                self._optidos_active = True
            else:
                self._logger.info("Unknown optiDos configuration / status")
        else:
            self._logger.info("optiDos is not active / available")

        self._logger.info("optiDos information: %s optiDos A status: %s, optiDos B status: %s",
                          self._optidos_config, self.optidos_a_status, self.optidos_b_status)

    async def _do_consumption_details_request(self, command: str) -> str:

        url = self.get_command_url(ENDPOINT_HH, COMMAND_GET_COMMAND).update_query({'value': command})
        eco_json = await self.make_vzug_device_call_json(url)

        if 'value' in eco_json:
            return eco_json['value']
        else:
            self._logger.error('Error reading power and water consumption, no \'value\' entry found in response.')
            raise DeviceError('Got invalid response while reading power and water consumption data.', 'n/a')

    async def load_consumption_data(self) -> bool:
        """Load power and water consumption data by calling the corresponding API endpoint"""

        self._logger.info("Loading power and water consumption data for %s", self._host)
        try:
            consumption_total = await self._do_consumption_details_request(COMMAND_VALUE_ECOM_STAT_TOTAL)
            self._power_consumption_kwh_total = self._read_kwh_from_string(consumption_total)
            self._water_consumption_l_total = self._read_liter_from_string(consumption_total)

            consumption_avg = await self._do_consumption_details_request(COMMAND_VALUE_ECOM_STAT_AVG)
            self._power_consumption_kwh_avg = self._read_kwh_from_string(consumption_avg)
            self._water_consumption_l_avg = self._read_liter_from_string(consumption_avg)

            self._logger.info("Power consumption total: %s kWh, avg: %.1f kWh",
                              locale.format_string('%.0f', self._power_consumption_kwh_total, True),
                              self._power_consumption_kwh_avg)

            self._logger.info("Water consumption total: %s l, avg: %.0f l",
                              locale.format_string('%.0f', self._water_consumption_l_total, True),
                              self._water_consumption_l_avg)

        except DeviceError as e:
            self._error_code = e.error_code
            self._error_message = e.message
            self._error_exception = e
            return False

        return True

    @staticmethod
    def _read_float_from_string(consumption_value: str, regex: str) -> float:
        match = re.search(regex, consumption_value)
        if match:
            return float(match.group(1).replace(',', '.'))

        return -1

    def _read_liter_from_string(self, consumption_value: str) -> float:
        liter = self._read_float_from_string(consumption_value, REGEX_MATCH_LITER)
        if liter < 0:
            raise DeviceError('Cannot find liter value in string %s'.format(consumption_value), 'n/a')

        return liter

    def _read_kwh_from_string(self, consumption_value: str) -> float:
        kwh = self._read_float_from_string(consumption_value, REGEX_MATCH_KWH)
        if kwh < 0:
            raise DeviceError('Cannot find kWh value in string %s'.format(consumption_value), 'n/a')

        return kwh

    @property
    def program_status(self) -> str:
        return self._program_status

    @property
    def program_name(self) -> str:
        return self._program_name

    @property
    def optidos_active(self) -> bool:
        return self._optidos_active

    @property
    def optidos_a_status(self) -> str:
        return self._optidos_a_status

    @property
    def optidos_b_status(self) -> str:
        return self._optidos_b_status

    @property
    def seconds_to_end(self) -> int:
        return self._seconds_to_end

    @property
    def date_time_end(self) -> datetime:
        return datetime.now() + timedelta(seconds=self.seconds_to_end)

    @property
    def power_consumption_kwh_total(self) -> float:
        return self._power_consumption_kwh_total

    @property
    def power_consumption_kwh_avg(self) -> float:
        return self._power_consumption_kwh_avg

    @property
    def water_consumption_l_total(self) -> float:
        return self._water_consumption_l_total

    @property
    def water_consumption_l_avg(self) -> float:
        return self._water_consumption_l_avg



from distutils.util import strtobool
import json

import aiohttp
import logging
from .const import (COMMAND_GET_STATUS, COMMAND_GET_MODEL_DESC, ENDPOINT_AI, VERSION)
from yarl import URL
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception_type, before_log

from .digest_auth import DigestAuth

REQUEST_HEADERS = {
    f"User-Agent": f"vzug-lib/{VERSION}",
    "Accept": f"application/json, text/plain, */*",
}


class DeviceError(Exception):
    def __init__(self, message, err_code: str, inner_exception: Exception = None):
        super().__init__(message)
        self._device_err_code = err_code
        self._message = message
        self._inner_exception = inner_exception

    @property
    def error_code(self) -> str:
        return self._device_err_code

    @property
    def message(self) -> str:
        return self._message

    @property
    def inner_exception(self) -> Exception:
        return self._inner_exception


class BasicDevice:
    """Class containing basic functions valid to any V-ZUG device"""

    def __init__(self, host: str, username: str = "", password: str = "") -> None:
        self._host = host
        self._username = username
        self._password = password
        self._serial = ""
        self._model_desc = ""
        self._device_name = ""
        self._status = ""
        self._status_json = json
        self._program = ""
        self._error_code = ""
        self._error_message = ""
        self._error_exception = None
        self._uuid = ""
        self._active = False
        self._device_information_loaded = False
        self._logger = logging.getLogger(__name__)

    def get_base_url(self) -> URL:
        return URL.build(scheme='http', host=self._host)

    def get_command_url(self, endpoint: str, command: str) -> URL:
        return self.get_base_url().join(URL(endpoint)).update_query({'command': command})

    async def make_vzug_device_call_raw(self, url: URL) -> str:
        """
        Make raw service call to any V-Zug device and return the response as text
        """

        async with aiohttp.ClientSession() as session:
            try:
                self._logger.debug("Raw service call URL: %s", str(url))

                auth = DigestAuth(self._username, self._password, session)
                resp = await auth.request('GET', url=url, headers=REQUEST_HEADERS)

                txt_resp = await resp.read()
                self._logger.debug("Raw response from %s: status %s, text: %s", self._host, resp.status, txt_resp)
                return txt_resp.decode("utf-8")

            except IOError as e:
                err_msg = "IOError while calling device API"
                self._logger.error("%s: %s", err_msg, str(e))
                raise DeviceError(err_msg, "n/a", e)

    @retry(stop=stop_after_attempt(3),
           wait=wait_fixed(2),
           retry=retry_if_exception_type(DeviceError),
           before=before_log(logging.getLogger(__name__), logging.DEBUG),
           reraise=True)
    async def make_vzug_device_call_json(self, url: URL) -> json:
        """
        Make service call for any V-Zug device and check if there is an error code in json response.
        Sometimes the devices returns an internal error (like 503). In this case DeviceError exception
        is raised after 3 retries.
        """

        try:
            text_resp = str(await self.make_vzug_device_call_raw(url))

            json_resp = json.loads(text_resp)
            if "error" in json_resp:
                err_code = json_resp['error']['code']
                self._logger.error("Device returned error code: %s", err_code)
                raise DeviceError("Device returned error code", err_code)
            return json_resp

        except ValueError as e:
            err_msg = "Got invalid response from device"
            self._logger.error("%s: %s", err_msg, str(e))
            raise DeviceError(err_msg, "n/a", e)

    async def load_all_information(self) -> bool:
        """
        For the basic device forward the call to load_device_information().
        Method can be overridden by subclasses to load device specific
        information.
        """
        return await self.load_device_information()

    async def load_device_information(self) -> bool:
        """Load device status information by calling the corresponding API endpoint"""

        try:
            self._logger.info("Loading device information for %s", self._host)
            self._status_json = await self.make_vzug_device_call_json(
                self.get_command_url(ENDPOINT_AI, COMMAND_GET_STATUS))

            self._error_code = ""
            self._serial = self._status_json['Serial']
            self._device_name = self._status_json['DeviceName']
            self._status = self._status_json['Status']
            self._uuid = self._status_json['deviceUuid']
            self._program = self._status_json['Program']
            self._active = not strtobool(self._status_json['Inactive'])

            # Load model description in separate call
            self._model_desc = await self.make_vzug_device_call_raw(
                self.get_command_url(ENDPOINT_AI, COMMAND_GET_MODEL_DESC))

            self._device_information_loaded = True

            self._logger.info("Go device information. Model: %s, Serial: %s, Uuid: %s, name: %s, Status text: %s",
                              self._model_desc, self._serial, self._uuid, self._device_name, self._status)
            return True

        except DeviceError as e:
            self._error_code = e.error_code
            self._error_message = e.message
            self._error_exception = e
            return False

    @property
    def serial(self) -> str:
        return self._serial

    @property
    def device_name(self) -> str:
        return self._device_name

    @property
    def model_desc(self) -> str:
        return self._model_desc

    @property
    def status(self) -> str:
        return self._status

    @property
    def status_json(self) -> json:
        return self._status_json

    @property
    def is_active(self) -> bool:
        return self._active

    @property
    def program(self) -> str:
        return self._program

    @property
    def error_code(self) -> str:
        return self._error_code

    @property
    def error_message(self) -> str:
        return self._error_message

    @property
    def error_exception(self) -> DeviceError:
        return self._error_exception

    @property
    def device_information_loaded(self) -> bool:
        return self._device_information_loaded

    @property
    def uuid(self) -> str:
        return self._uuid



VERSION = '0.0.3'

ENDPOINT_AI = 'ai'
ENDPOINT_HH = 'hh'
COMMAND_GET_STATUS = 'getDeviceStatus'
COMMAND_GET_MODEL_DESC = 'getModelDescription'
COMMAND_GET_PROGRAM = 'getProgram'
COMMAND_GET_COMMAND = 'getCommand'



from aiohttp import client_exceptions, hdrs
import hashlib
from yarl import URL
import time
import os


class DigestAuth:
    """HTTP digest authentication helper.
    The work here is based off of
    https://github.com/requests/requests/blob/v2.18.4/requests/auth.py.
    """

    def __init__(self, username, password, session, previous=None):
        if previous is None:
            previous = {}

        self.username = username
        self.password = password
        self.last_nonce = previous.get('last_nonce', '')
        self.nonce_count = previous.get('nonce_count', 0)
        self.challenge = previous.get('challenge')
        self.args = {}
        self.session = session

    async def request(self, method, url, *, headers=None, **kwargs):
        if headers is None:
            headers = {}

        # Save the args so we can re-run the request
        self.args = {
            'method': method,
            'url': url,
            'headers': headers,
            'kwargs': kwargs
        }

        if self.challenge:
            headers[hdrs.AUTHORIZATION] = self._build_digest_header(
                method.upper(), url
            )

        response = await self.session.request(
            method, url, headers=headers, **kwargs
        )

        # Only try performing digest authentication if the response status is
        # from 400 to 500.
        if 400 <= response.status < 500:
            return await self._handle_401(response)

        return response

    def _build_digest_header(self, method, url):
        """
        :rtype: str
        """

        realm = self.challenge['realm']
        nonce = self.challenge['nonce']
        qop = self.challenge.get('qop')
        algorithm = self.challenge.get('algorithm', 'MD5').upper()
        opaque = self.challenge.get('opaque')

        if qop and not (qop == 'auth' or 'auth' in qop.split(',')):
            raise client_exceptions.ClientError(
                'Unsupported qop value: %s' % qop
            )

        # lambdas assume digest modules are imported at the top level
        if algorithm == 'MD5' or algorithm == 'MD5-SESS':
            hash_fn = hashlib.md5
        elif algorithm == 'SHA':
            hash_fn = hashlib.sha1
        else:
            return ''

        def H(x):
            return hash_fn(x.encode()).hexdigest()

        def KD(s, d):
            return H('%s:%s' % (s, d))

        path = URL(url).path_qs
        A1 = '%s:%s:%s' % (self.username, realm, self.password)
        A2 = '%s:%s' % (method, path)

        HA1 = H(A1)
        HA2 = H(A2)

        if nonce == self.last_nonce:
            self.nonce_count += 1
        else:
            self.nonce_count = 1

        self.last_nonce = nonce

        ncvalue = '%08x' % self.nonce_count

        # cnonce is just a random string generated by the client.
        cnonce_data = ''.join([
            str(self.nonce_count),
            nonce,
            time.ctime(),
            os.urandom(8).decode(errors='ignore'),
        ]).encode()
        cnonce = hashlib.sha1(cnonce_data).hexdigest()[:16]

        if algorithm == 'MD5-SESS':
            HA1 = H('%s:%s:%s' % (HA1, nonce, cnonce))

        # This assumes qop was validated to be 'auth' above. If 'auth-int'
        # support is added this will need to change.
        if qop:
            noncebit = ':'.join([
                nonce, ncvalue, cnonce, 'auth', HA2
            ])
            response_digest = KD(HA1, noncebit)
        else:
            response_digest = KD(HA1, '%s:%s' % (nonce, HA2))

        base = ', '.join([
            'username="%s"' % self.username,
            'realm="%s"' % realm,
            'nonce="%s"' % nonce,
            'uri="%s"' % path,
            'response="%s"' % response_digest,
            'algorithm="%s"' % algorithm,
        ])
        if opaque:
            base += ', opaque="%s"' % opaque
        if qop:
            base += ', qop="auth", nc=%s, cnonce="%s"' % (ncvalue, cnonce)

        return 'Digest %s' % base

    async def _handle_401(self, response):
        """
        Takes the given response and tries digest-auth, if needed.
        :rtype: ClientResponse
        """
        auth_header = response.headers.get('www-authenticate', '')

        parts = auth_header.split(' ', 1)
        if 'digest' == parts[0].lower() and len(parts) > 1:
            self.challenge = parse_key_value_list(parts[1])

            return await self.request(
                self.args['method'],
                self.args['url'],
                headers=self.args['headers'],
                **self.args['kwargs']
            )

        return response


def parse_pair(pair):
    key, value = pair.split('=', 1)

    # If it has a trailing comma, remove it.
    if value[-1] == ',':
        value = value[:-1]

    # If it is quoted, then remove them.
    if value[0] == value[-1] == '"':
        value = value[1:-1]

    return key, value


def parse_key_value_list(header):
    return {
        key: value for key, value in
        [parse_pair(header_pair) for header_pair in header.split(' ')]
    }



import os
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, "README_PYPI.md"), encoding="utf-8") as readme:
    long_desc = readme.read()

setup(
    name='vzug-api',
    version='0.0.3',
    description='Unofficial python API for V-ZUG devices',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url='https://github.com/mico-micic/vzug-api',
    author='Mićo Mićić',
    author_email='mico@micic.ch',
    license='GNU General Public License v3.0',
    packages=find_packages(exclude=["test", "examples"]),
    include_package_data=True,
    install_requires=[
        'aiohttp>=3.8.0',
        'tenacity>=8.0.0',
        'yarl>=1.7.0'
    ],
    tests_require=['pytest'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Topic :: Home Automation",
    ],
)
